#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <unistd.h>

#include <string>

extern int	errno;

/*
 *  Global variables
 */

static char *usageStr = " -t <tunnel port>";
static int  debug = 0;

static int			listenSock;
static int			tcpSock;

int getline(int fd, string &str)
{
    char		c;
    int		nBytes;

    str = "";
    do
    {
	nBytes = read(fd, &c, 1);
	if ( nBytes != 1 )
	    return 1;

	if ( c != '\n' )
	    str += c;
    }  while ( c != '\n' );

    return 0;
}

void tunnelEnd(int socket)
{	
    string 	str;
    int nBytes = 0;
    fd_set 	read_fd;
    fd_set 	error_fd;
    
    FD_ZERO(&read_fd);
    FD_ZERO(&error_fd);
    
    for (;;)
    {
	FD_SET(0, &read_fd);
	FD_SET(socket, &read_fd);
	FD_SET(socket, &error_fd);

	select(32, &read_fd, NULL, &error_fd, NULL);

	// 1st, Handle error situations
	if ( FD_ISSET(socket, &error_fd) )
	{
	    fprintf(stderr, "Socket error: %d\n", errno);
	    close (socket);
	    return;
	}

	// Then copy to socket what we've got from stdin
	if ( FD_ISSET(0, &read_fd) )
	{
	    getline(0,str);
	    str += '\n';
	    write(socket,str.c_str(),strlen(str.c_str()));
	      
	}

	// and vice-versa
	if ( FD_ISSET(socket, &read_fd) )
	{
	    if ( getline(socket,str) != 0 )
	    {
		fprintf(stderr, "Getline returned 0, %s (%d)\n", strerror(errno), errno);
		close (socket);
		return;
	    }
	    fprintf(stdout, "%s\n", str.c_str());
	}
    }
}
    
void waitForConnection (void)
    {
    int			addrLen;
    int			tcpSock;
    int			childPid;
    struct sockaddr_in	peerAddr;
    
    for (;;)
	{
	if ( debug )
	    fprintf(stderr, "Ready to accept new connections\n");
	
	memset(&peerAddr, 0, sizeof(struct sockaddr_in));
	addrLen = sizeof(struct sockaddr_in);
	
	tcpSock = accept(listenSock, (struct sockaddr *) &peerAddr, &addrLen);
	if ( tcpSock == -1 )
	    fprintf (stderr, "accept() failed\n");
	
	if ( debug )
	    {
	    fprintf(stderr, "Received incoming connection from %s port %d\n",
		    inet_ntoa(peerAddr.sin_addr),
		    peerAddr.sin_port
		    );
	    }

	tunnelEnd (tcpSock);
	sleep(1);
 	}
    }

main(int argc, char *argv[])
    {
    struct sockaddr_in	locTcpSockAddr;
    struct hostent 	*hp;
    int			tunnelPortNum = 1199;
    int			rc;
    char		c;
    extern char * optarg;

    /* Decode input args */
    while ( (c = getopt(argc, argv, "dt:")) != -1 )
	switch ( c )
	    {
	case 'd':
	    debug = 1;
	    break;
	    
	case 't' :
	    tunnelPortNum = atoi(optarg);
	    break;

	case '?' :
	    fprintf(stderr,usageStr);
	    exit (1);
	    }

    /* Misc. checks */
    
    if ( tunnelPortNum == -1 )
	{
	fprintf(stderr, "%s: -t <port num> arguments missing\n", argv[0]);
	exit (1);
	}
    
    /* Reset structures */
    memset((char *) &locTcpSockAddr, 0, sizeof(struct sockaddr_in));
	
    /* Set up address structure for TCP listening socket */
    locTcpSockAddr.sin_family = AF_INET;
    locTcpSockAddr.sin_addr.s_addr = INADDR_ANY;
    locTcpSockAddr.sin_port = htons(tunnelPortNum);

    /* Create the TCP listen socket */
    listenSock = socket (AF_INET, SOCK_STREAM, 0);
    if (listenSock == -1)
	/* printSysErr("TCP listen socket creation"); */
	{
	fprintf(stderr, "%s: TCP listen socket creation\n", argv[0]);
	exit (1);
	}
	

    /* Bind the listening address to the socket */
    if ( bind(listenSock, (struct sockaddr *) &locTcpSockAddr, sizeof(struct sockaddr_in)) == -1 )
      {
	fprintf(stderr, "%s: TCP socket bind\n", argv[0]);
	exit (1);
      }

    /* Listen for incoming connections */
    if ( listen(listenSock, 5) == - 1)
      {
	fprintf(stderr, "%s: listen() failed\n", argv[0]);
	exit (1);
      }

    waitForConnection();
    
    exit (0);
    }
